@@ -1,7 +1,7 @@
 package JobCenter::Client::Mojo;
 use Mojo::Base -base;
 
-our $VERSION = '0.33'; # VERSION
+our $VERSION = '0.34'; # VERSION
 
 #
 # Mojo's default reactor uses EV, and EV does not play nice with signals
@@ -11,6 +11,7 @@ our $VERSION = '0.33'; # VERSION
 BEGIN {
 	$ENV{'MOJO_REACTOR'} = 'Mojo::Reactor::Poll' unless $ENV{'MOJO_REACTOR'};
 }
+
 # more Mojolicious
 use Mojo::IOLoop;
 use Mojo::IOLoop::Stream;
@@ -39,6 +40,13 @@ has [qw(
 	log method ns ping_timeout port rpc timeout tls token who
 )];
 
+# keep in sync with the jobcenter
+use constant {
+	WORK_OK                => 0,  # exit codes for work method
+	WORK_PING_TIMEOUT      => 92,
+	WORK_CONNECTION_CLOSED => 91,
+};
+
 sub new {
 	my ($class, %args) = @_;
 	my $self = $class->SUPER::new();
@@ -57,27 +65,60 @@ sub new {
 	my $token = $args{token} or croak 'no token?';
 	my $who = $args{who} or croak 'no who?';
 
-	my $rpc = JSON::RPC2::TwoWay->new(debug => $debug) or croak 'no rpc?';
+	$self->{address} = $address;
+	$self->{daemon} = $args{daemon} // 0;
+	$self->{debug} = $args{debug} // 1;
+	$self->{jobs} = {};
+	$self->{json} = $json;
+	$self->{ping_timeout} = $args{ping_timeout} // 300;
+	$self->{log} = $log;
+	$self->{method} = $method;
+	$self->{port} = $port;
+	$self->{timeout} = $timeout;
+	$self->{tls} = $tls;
+	$self->{tls_ca} = $tls_ca;
+	$self->{tls_cert} = $tls_cert;
+	$self->{tls_key} = $tls_key;
+	$self->{token} = $token;
+	$self->{who} = $who;
+	$self->{autoconnect} = ($args{autoconnect} //= 1);
+
+	return $self if !$args{autoconnect};
+
+	$self->connect;
+
+	return $self if $self->{auth};
+	return;
+}
+
+sub connect {
+	my $self = shift;
+
+	delete $self->{_exit};
+	delete $self->{auth};
+	$self->{actions} = {};
+
+	my $rpc = JSON::RPC2::TwoWay->new(debug => $self->{debug}) or croak 'no rpc?';
 	$rpc->register('greetings', sub { $self->rpc_greetings(@_) }, notification => 1);
 	$rpc->register('job_done', sub { $self->rpc_job_done(@_) }, notification => 1);
 	$rpc->register('ping', sub { $self->rpc_ping(@_) });
 	$rpc->register('task_ready', sub { $self->rpc_task_ready(@_) }, notification => 1);
 
 	my $clarg = {
-		address => $address,
-		port => $port,
-		tls => $tls,
+		address => $self->{address},
+		port => $self->{port},
+		tls => $self->{tls},
 	};
-	$clarg->{tls_ca} = $tls_ca if $tls_ca;
-	$clarg->{tls_cert} = $tls_cert if $tls_cert;
-	$clarg->{tls_key} = $tls_key if $tls_key;
+	$clarg->{tls_ca} = $self->{tls_ca} if $self->{tls_ca};
+	$clarg->{tls_cert} = $self->{tls_cert} if $self->{tls_cert};
+	$clarg->{tls_key} = $self->{tls_key} if $self->{tls_key};
 
 	my $clientid = Mojo::IOLoop->client(
 		$clarg => sub {
 		my ($loop, $err, $stream) = @_;
 		if ($err) {
 			$err =~ s/\n$//s;
-			$log->info('connection to API failed: ' . $err);
+			$self->log->info('connection to API failed: ' . $err);
 			$self->{auth} = 0;
 			return;
 		}
@@ -92,53 +133,42 @@ sub new {
 			my ($ns2, $chunk) = @_;
 			#say 'got chunk: ', $chunk;
 			my @err = $conn->handle($chunk);
-			$log->debug('chunk handler: ' . join(' ', grep defined, @err)) if @err;
+			$self->log->debug('chunk handler: ' . join(' ', grep defined, @err)) if @err;
 			$ns->close if $err[0];
 		});
 		$ns->on(close => sub {
 			$conn->close;
-			$log->info('connection to API closed');
-			$self->{_exit} = 91; # todo: doc
+			$self->log->info('connection to API closed');
+			$self->{_exit} = WORK_CONNECTION_CLOSED; # todo: doc
 			Mojo::IOLoop->stop;
 			#exit(1);
 		});
 	});
 
-	$self->{actions} = {};
-	$self->{address} = $address;
-	$self->{clientid} = $clientid;
-	$self->{daemon} = $args{daemon} // 0;
-	$self->{debug} = $args{debug} // 1;
-	$self->{jobs} = {};
-	$self->{json} = $json;
-	$self->{ping_timeout} = $args{ping_timeout} // 300;
-	$self->{log} = $log;
-	$self->{method} = $method;
-	$self->{port} = $port;
 	$self->{rpc} = $rpc;
-	$self->{timeout} = $timeout;
-	$self->{tls} = $tls;
-	$self->{tls_ca} = $tls_ca;
-	$self->{tls_cert} = $tls_cert;
-	$self->{tls_key} = $tls_key;
-	$self->{token} = $token;
-	$self->{who} = $who;
+	$self->{clientid} = $clientid;
 
 	# handle timeout?
-	my $tmr = Mojo::IOLoop->timer($timeout => sub {
+	my $tmr = Mojo::IOLoop->timer($self->{timeout} => sub {
 		my $loop = shift;
-		$log->error('timeout wating for greeting');
+		$self->log->error('timeout wating for greeting');
 		$loop->remove($clientid);
 		$self->{auth} = 0;
 	});
 
 	$self->log->debug('starting handshake');
-	Mojo::IOLoop->singleton->reactor->one_tick while !defined $self->{auth};
+
+	$self->_loop(sub { not defined $self->{auth} });
+
 	$self->log->debug('done with handhake?');
 
 	Mojo::IOLoop->remove($tmr);
-	return $self if $self->{auth};
-	return;
+	1;
+}
+
+sub is_connected {
+	my $self = shift;
+	return $self->{auth} && !$self->{_exit};
 }
 
 sub rpc_greetings {
@@ -205,7 +235,7 @@ sub call {
 	};
 	$self->call_nb(%args);
 
-	Mojo::IOLoop->singleton->reactor->one_tick while !$done;
+	$self->_loop(sub { !$done });
 
 	return $job_id, $outargs;
 }
@@ -322,7 +352,7 @@ sub find_jobs {
 		$done++;
 	});
 
-	Mojo::IOLoop->singleton->reactor->one_tick while !$done;
+	$self->_loop(sub { !$done });
 
 	return $err, @$jobs if ref $jobs eq 'ARRAY';
 	return $err;
@@ -384,7 +414,7 @@ sub get_job_status {
 		$done++;
 	});
 
-	Mojo::IOLoop->singleton->reactor->one_tick while !$done;
+	$self->_loop(sub { !$done });
 
 	$outargs = encode_json($outargs) if $self->{json} and ref $outargs;
 	return $job_id2, $outargs;
@@ -410,7 +440,7 @@ sub ping {
 		$done++;
 	});
 
-	Mojo::IOLoop->singleton->reactor->one_tick while !$done;
+	$self->_loop(sub { !$done });
 
 	return $ret;
 }
@@ -422,21 +452,24 @@ sub work {
 	}
 
 	my $pt = $self->ping_timeout;
-	my $tmr = Mojo::IOLoop->recurring($pt => sub {
+	my $tmr;
+	$tmr = Mojo::IOLoop->recurring($pt => sub {
 		my $ioloop = shift;
 		$self->log->debug('in ping_timeout timer: lastping: '
 			 . ($self->lastping // 0) . ' limit: ' . (time - $pt) );
 		return if ($self->lastping // 0) > time - $pt;
 		$self->log->error('ping timeout');
 		$ioloop->remove($self->clientid);
-		$self->{_exit} = 92; # todo: doc
+		$self->{_exit} = WORK_PING_TIMEOUT; # todo: doc
+		$ioloop->remove($tmr);
 		$ioloop->stop;
 	}) if $pt > 0;
 
-	$self->{_exit} = 0;
+	$self->{_exit} = WORK_OK;
 	$self->log->debug('JobCenter::Client::Mojo starting work');
 	Mojo::IOLoop->start unless Mojo::IOLoop->is_running;
 	$self->log->debug('JobCenter::Client::Mojo done?');
+	Mojo::IOLoop->remove($tmr) if $tmr;
 
 	return $self->{_exit};
 }
@@ -643,6 +676,36 @@ sub _daemonize {
 	open STDERR, '>&STDOUT';
 }
 
+# tick while Mojo::Reactor is still running and condition callback is true
+sub _loop {
+	warn __PACKAGE__." recursing into IO loop" if state $looping++;
+
+	my $reactor = Mojo::IOLoop->singleton->reactor;
+	my $err;
+
+	if (ref $reactor eq 'Mojo::Reactor::EV') {
+
+		my $active = 1;
+
+		$active = $reactor->one_tick while $_[1]->() && $active;
+
+	} elsif (ref $reactor eq 'Mojo::Reactor::Poll') {
+
+		$reactor->{running}++;
+
+		$reactor->one_tick while $_[1]->() && $reactor->is_running;
+
+		$reactor->{running} &&= $reactor->{running} - 1;
+
+	} else {
+
+		$err = "unknown reactor: ".ref $reactor;
+	}
+
+	$looping--;
+	die $err if $err;
+}
+
 #sub DESTROY {
 #	my $self = shift;
 #	say 'destroying ', $self;
@@ -926,7 +989,9 @@ callback mode the environment will be inserted before the result callback.
 
 =head2 work
 
-Starts the L<Mojo::IOLoop>.
+Starts the L<Mojo::IOLoop>.  Returns a non-zero value when the IOLoop was
+stopped due to some error condition (like a lost connection or a ping
+timeout).
 
 =head1 SEE ALSO
 
